//
//  CCCloudManager.m
//  iCloudDemoMOBC2013
//
//  Created by Gianluca Tranchedone on 20/10/2013.
//  Copyright (c) 2013 Cocoa Beans GT Limited. All rights reserved.
//

#import "CCCloudManager.h"

@implementation CCCloudManager

@synthesize managedObjectModel = _managedObjectModel;
@synthesize managedObjectContext = _managedObjectContext;
@synthesize persistentStoreCoordinator = _persistentStoreCoordinator;

#pragma mark - Superclass Methods Override

- (id)init
{
    self = [super init];
    if (self) {
        [self setupCoreDataStack];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(persistentStoreWillChange:)
                                                     name:NSPersistentStoreCoordinatorStoresWillChangeNotification
                                                   object:self.persistentStoreCoordinator];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(persistentStoreDidChange:)
                                                     name:NSPersistentStoreCoordinatorStoresDidChangeNotification
                                                   object:self.persistentStoreCoordinator];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(persistentStoreDidImportChanges:)
                                                     name:NSPersistentStoreDidImportUbiquitousContentChangesNotification
                                                   object:nil];
    }
    return self;
}

#pragma mark - Public APIs

+ (instancetype)sharedManager
{
    static CCCloudManager *sharedManager = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        sharedManager = [[self alloc] init];
    });
    return sharedManager;
}

+ (NSURL *)applicationDocumentsDirectory
{
    return [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask] lastObject];
}

- (void)saveContext
{
    [self saveManagedObjectContext:self.managedObjectContext asynchronously:NO];
    [self saveManagedObjectContext:self.managedObjectContext.parentContext asynchronously:YES];
}

#pragma mark - Private APIs

- (void)setupCoreDataStack
{
    NSPersistentStoreCoordinator *storeCoordinator = [self persistentStoreCoordinator];
    if (storeCoordinator) {
        NSMergePolicy *mergePolicy = [[NSMergePolicy alloc] initWithMergeType:NSMergeByPropertyStoreTrumpMergePolicyType];
        
        NSManagedObjectContext *savingContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];
        savingContext.persistentStoreCoordinator = storeCoordinator;
        savingContext.mergePolicy = mergePolicy;
        
        NSManagedObjectContext *managedObjectContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSMainQueueConcurrencyType];
        managedObjectContext.parentContext = savingContext;
        managedObjectContext.mergePolicy = mergePolicy;
        
        _managedObjectContext = managedObjectContext;
    }
}

- (void)persistentStoreWillChange:(NSNotification *)notification
{
    [self saveContext];
    [self.managedObjectContext reset];
}

- (void)persistentStoreDidChange:(NSNotification *)notification
{
    // TODO: refresh the UI if necessary
}

- (void)persistentStoreDidImportChanges:(NSNotification *)notification
{
    if ([NSThread isMainThread]) {
        [self.managedObjectContext mergeChangesFromContextDidSaveNotification:notification];
        [self saveContext];
    }
    else {
        dispatch_async(dispatch_get_main_queue(), ^{
            [self.managedObjectContext mergeChangesFromContextDidSaveNotification:notification];
            [self saveContext];
        });
    }
}

- (void)saveManagedObjectContext:(NSManagedObjectContext *)managedObjectContext asynchronously:(BOOL)asynchronously
{
    __weak NSManagedObjectContext *contextToSave = managedObjectContext;
    void (^saveBlock)(void) = ^{
        NSError *error = nil;
        NSManagedObjectContext *context = contextToSave; // get a strong pointer again
        
        if ([context hasChanges] && ![context save:&error]) {
            // Replace this implementation with code to handle the error appropriately.
            // abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development.
            NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
            abort();
        }
    };
    
    if (asynchronously) {
        [contextToSave performBlock:saveBlock];
    }
    else {
        [contextToSave performBlockAndWait:saveBlock];
    }
}

#pragma mark - Setters and Getters

- (NSManagedObjectModel *)managedObjectModel
{
    if (!_managedObjectModel) {
        NSURL *modelURL = [[NSBundle mainBundle] URLForResource:@"iCloudDemoMOBC2013" withExtension:@"momd"];
        _managedObjectModel = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];
    }
    
    return _managedObjectModel;
}

- (NSPersistentStoreCoordinator *)persistentStoreCoordinator
{
    if (!_persistentStoreCoordinator) {
        NSURL *storeURL = [[self.class applicationDocumentsDirectory] URLByAppendingPathComponent:@"iCloudDemoMOBC2013.sqlite"];
        
        NSError *error = nil;
        NSDictionary *options = @{NSPersistentStoreUbiquitousContentNameKey: @"iCloudDemoMOBC2013",
                                  NSMigratePersistentStoresAutomaticallyOption: @YES,
                                  NSInferMappingModelAutomaticallyOption: @YES};
        
        _persistentStoreCoordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:self.managedObjectModel];
        if (![_persistentStoreCoordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storeURL options:options error:&error]) {
            /*
             Replace this implementation with code to handle the error appropriately.
             
             abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development.
             
             Typical reasons for an error here include:
             * The persistent store is not accessible;
             * The schema for the persistent store is incompatible with current managed object model.
             Check the error message to determine what the actual problem was.
             
             
             If the persistent store is not accessible, there is typically something wrong with the file path. Often, a file URL is pointing into the application's resources directory instead of a writeable directory.
             
             If you encounter schema incompatibility errors during development, you can reduce their frequency by:
             * Simply deleting the existing store:
             [[NSFileManager defaultManager] removeItemAtURL:storeURL error:nil]
             
             * Performing automatic lightweight migration by passing the following dictionary as the options parameter:
             @{NSMigratePersistentStoresAutomaticallyOption:@YES, NSInferMappingModelAutomaticallyOption:@YES}
             
             Lightweight migration will only work for a limited set of schema changes; consult "Core Data Model Versioning and Data Migration Programming Guide" for details.
             
             */
            NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
            abort();
        }
    }
    
    return _persistentStoreCoordinator;
}

@end
